package cn.jisol.game.actions.onhook;

import cn.hutool.json.JSON;
import cn.hutool.json.JSONUtil;
import cn.jisol.game.actions.GActionEnum;
import cn.jisol.game.actions.GActionEvent;
import cn.jisol.game.actions.GRefereeAction;
import cn.jisol.game.controller.game.GPlayerPetController;
import cn.jisol.game.controller.game.GPlayerTacticalController;
import cn.jisol.game.entity.game.PlayerTactical;
import cn.jisol.game.network.client.GClient;
import cn.jisol.game.network.client.GRefereeClient;
import cn.jisol.game.proto.GPVPMessage;
import cn.jisol.game.proto.GUIMessage;
import cn.jisol.game.service.PlayerTacticalService;
import cn.jisol.ngame.actions.core.NAction;
import cn.jisol.ngame.actions.core.NActionMethod;
import cn.jisol.ngame.util.EventDispatcher;
import cn.jisol.ngame.util.spring.SpringBeanUtils;
import java.util.*;

//PVP 模式
@NAction
public class GPVPAction {

    //当前等待PVP的玩家
    static LinkedList<GClient> pool = new LinkedList<>();

    //等待通知定时器
    static Timer timer;

    static {

        timer = new Timer();
        timer.schedule(new TimerTask() {
            @Override
            public void run() {
                GPVPAction.onWaitTimer();
            }
        },0,1000);

        EventDispatcher.getInstance().addListener(GActionEvent.REFEREE_FREE,(EventDispatcher.Event event) -> {
            onUpdateMatchGame(); //刷新匹配
        });

    }

    //加入PVP
    @NActionMethod(GActionEnum.S_MODE_PVP_JOIN)
    public static synchronized void onJoinPVP(GClient client){

        //如果加入过 PVP 则返回
        if(pool.contains(client)) return;

        for (int i = 0; i < pool.size(); i++) {
            GClient item = pool.get(i);
            if (item.player.getPlayerId().equals(client.player.getPlayerId())){
                //移除玩家
                pool.remove(item);
            }
        }

        //如果没有加入则加入
        pool.add(client);

        //通知开始等待
        client.invoke(GActionEnum.C_MODE_PVP_START_WAIT);

        GPVPAction.onUpdateMatchGame();

    }

    //离开PVP
    @NActionMethod(GActionEnum.S_MODE_PVP_LEAVE)
    public static void onJoinLeave(GClient client){

        //如果加入过 则 离开
        if(pool.contains(client)) {
            pool.remove(client);
            client.invoke(GActionEnum.C_MODE_PVP_END_WAIT);
        }

    }

    //刷新匹配
    public static void onUpdateMatchGame(){

        //判断是否有两人在等待 有 则两人匹配成功
        while (pool.size() >= 2){

            if (!(GRefereeAction.autoReferee(referee -> {

                //匹配成功
                GClient client1 = pool.poll();
                GClient client2 = pool.poll();

                if(Objects.isNull(client1) || Objects.isNull(client2)) {
                    return referee;
                }
                //确保玩家都在线
                if(!(client1.isOpen() &&  client2.isOpen())){
                    if(client2.isOpen()) pool.addFirst(client2);
                    if(client1.isOpen()) pool.addFirst(client1);
                    return referee;
                }

                //获取双方的阵型
                GPlayerTacticalController tacticalController = SpringBeanUtils.getBean(GPlayerTacticalController.class);
                GPlayerPetController petController = SpringBeanUtils.getBean(GPlayerPetController.class);

                String leftTactical = tacticalController.getInfo(client1.player).data.getTacticalData();
                String rightTactical = tacticalController.getInfo(client2.player).data.getTacticalData();


                //如果有人阵法是默认 则 不开始
                if (leftTactical.equals(tacticalController.InitTactical) || rightTactical.equals(tacticalController.InitTactical)){
                    if(leftTactical.equals(tacticalController.InitTactical)){
                        client1.invoke(GActionEnum.C_MODE_PVP_END_WAIT);
                        client1.invoke(GActionEnum.C_MODE_PVP_MESSAGE, GPVPMessage.GPVPText.newBuilder().setText("你的阵法没有宠物~").build());
                    }else{
                        //加入回列表
                        pool.addFirst(client1);
                    }
                    if(rightTactical.equals(tacticalController.InitTactical)){
                        client2.invoke(GActionEnum.C_MODE_PVP_END_WAIT);
                        client2.invoke(GActionEnum.C_MODE_PVP_MESSAGE, GPVPMessage.GPVPText.newBuilder().setText("你的阵法没有宠物~").build());
                    }else{
                        //加入回列表
                        pool.addFirst(client2);
                    }
                    return referee;
                }

                //构建匹配信息
                GPVPMessage.GPVPStart.Builder builder = GPVPMessage.GPVPStart.newBuilder()
                        .setLeftPlayerId(client1.player.getPlayerId())
                        .setRightPlayerId(client2.player.getPlayerId())
                        .setLeftTactical(leftTactical)
                        .setRightTactical(rightTactical);
                //构建玩家宠物列表
                petController.getPetList(client1.player).data.forEach(pet -> {
                    builder.putLeftPets(pet.getPetId(), JSONUtil.toJsonStr(pet));
                });
                petController.getPetList(client2.player).data.forEach(pet -> {
                    builder.putLeftPets(pet.getPetId(), JSONUtil.toJsonStr(pet));
                });

                //取消等待
                client1.invoke(GActionEnum.C_MODE_PVP_END_WAIT);
                client2.invoke(GActionEnum.C_MODE_PVP_END_WAIT);

                GPVPMessage.GPVPStart info = builder.build();

                //PVP 开始
                client1.invoke(GActionEnum.C_MODE_PVP_START, info);
                client2.invoke(GActionEnum.C_MODE_PVP_START, info);

                //PVP 裁决
                onPVPReferee(referee,info);

                return null; //返回空则占用 需要 手动释放
            }))){
                System.out.println("匹配失败 暂无裁决者空闲");
                //如果获取不到裁决 则 返回
                return;
            }

        }

    }

    //裁决PVP
    public static void onPVPReferee(GRefereeClient referee,GPVPMessage.GPVPStart info){
        //如果有裁决员则向裁决员发送要裁决的信息
        referee.invoke(GActionEnum.CR_REFEREE_PVP_MODE,info);
    }

    //裁决结束返回
    @NActionMethod(GActionEnum.CR_REFEREE_PVP_END)
    public static void onPVPRefereeEnd(GPVPMessage.GPVPRefereeInfo info,GRefereeClient referee,Map<String, GClient> clients){

        System.out.println("获胜的玩家是:"+info.getWinnerId());
        //裁决结束 等待下次裁决
        GRefereeAction.addFreeReferee(referee);


        clients.values().forEach(client -> {
            client.invoke(GActionEnum.C_MODE_PVP_MESSAGE, GPVPMessage.GPVPText.newBuilder().setText("服务器裁决胜利者"+info.getWinnerId()).build());
            client.invoke(GActionEnum.CHAT_RECEIVE_MESSAGE, GUIMessage.GUIChatMessage.newBuilder().setMessage("获胜的玩家是:"+info.getWinnerId()).build());
        });

    }

    //通知玩家等待定时器
    public static void onWaitTimer(){
        //向所有玩家发送等待
        pool.forEach(client -> {
            client.invoke(GActionEnum.C_MODE_PVP_WAIT);
        });
    }

}
