package cn.jisol.game.controller.game.mode;

import cfg.TB.TbGOnHookMaps;
import cfg.TB.TbGRole;
import cn.hutool.core.util.RandomUtil;
import cn.jisol.game.controller.argsresolver.CurrentPlayer;
import cn.jisol.game.data.GlobalId;
import cn.jisol.game.data.ResourceId;
import cn.jisol.game.data.TD;
import cn.jisol.game.entity.game.Player;
import cn.jisol.game.entity.game.PlayerPet;
import cn.jisol.game.entity.game.Resource;
import cn.jisol.game.entity.mode.ModeOnHook;
import cn.jisol.game.proto.GOnHookMessage;
import cn.jisol.game.service.ModeOnHookService;
import cn.jisol.game.service.PlayerPetService;
import cn.jisol.game.service.ResourceService;
import cn.jisol.game.vo.news.NewsResource;
import cn.jisol.game.vo.news.ov.ResourceUpdateOV;
import cn.jisol.ngame.util.NewsContext;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiImplicitParams;
import io.swagger.annotations.ApiOperation;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpHeaders;
import org.springframework.http.ResponseEntity;
import org.springframework.retry.RetryException;
import org.springframework.retry.annotation.Retryable;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.*;

import java.util.*;
import java.util.stream.Collectors;

@Api(value = "JNGameDemo - API", tags = {"无尽模式 - API"})
@RestController
@RequestMapping("/game/mode/onHook")
@ResponseBody
public class GOnHookController {

    @Autowired
    PlayerPetService playerPetService;
    @Autowired
    ResourceService resourceService;
    @Autowired
    ModeOnHookService modeOnHookService;

    //生成的野怪列表
    public static Map<Long, Map<String, GOnHookMessage.GOnHookPet>> CREEPS = new HashMap<>();

    //获取玩家模式信息
    @ApiImplicitParams({})
    @ApiOperation(value = "生成野怪")
    @GetMapping("/info")
    @Retryable(value = RetryException.class)
    public NewsContext<ModeOnHook> getInfo(@CurrentPlayer Player player){

        ModeOnHook info = modeOnHookService.getById(player.getPlayerId());

        //如果没有信息则初始化默认信息
        if(Objects.isNull(info)){
            //获取第一个地图
            TbGOnHookMaps oneMapInfo = TD.DATA.getTbGOnHookMaps().getDataList().get(0);
            if(Objects.isNull(oneMapInfo)){
                return NewsResource.onFail("获取信息失败 没有地图信息");
            }
            modeOnHookService.save(ModeOnHook.builder()
                        .playerId(player.getPlayerId())
                        .onHookMap(oneMapInfo.id)
                    .build());
            //并且重试获取
            throw new RetryException("");
        }

        return NewsContext.onSuccess("获取成功",info);

    }

    //切换地图
    @ApiImplicitParams({})
    @ApiOperation(value = "切换地图")
    @PostMapping("/setMapId/{mapId}")
    public NewsContext<ModeOnHook> setMapId(@CurrentPlayer Player player,@PathVariable Integer mapId){
        //判断是否有这个地图
        TbGOnHookMaps map = TD.DATA.getTbGOnHookMaps().get(mapId);
        if (Objects.isNull(map)) NewsContext.onFail("切换地图失败 没有这个地图哦");
        ModeOnHook data = this.getInfo(player).data;
        data.setOnHookMap(mapId);
        modeOnHookService.saveOrUpdate(data);
        //切换地图 清理历史野怪
        CREEPS.put(player.getPlayerId(),new HashMap<>());
        return NewsContext.onSuccess("成功",data);
    }

    //生成野怪
    @ApiImplicitParams({})
    @ApiOperation(value = "生成野怪")
    @GetMapping("/onSpawnCreeps")
    public ResponseEntity<byte[]> onSpawnCreeps(@CurrentPlayer Player player){

        HttpHeaders headers = new HttpHeaders();
        headers.set("Content-Type", "application/json");

        //获取玩家生成的宠物
        Map<String, GOnHookMessage.GOnHookPet> creeps = CREEPS.get(player.getPlayerId());

        //如果没有则生成类
        if(Objects.isNull(creeps)){
            CREEPS.put(player.getPlayerId(),creeps = new HashMap<>());
        }

        //获取可生成的最大数量
        int max = Integer.parseInt(TD.DATA.getTbGOnHookGlobal().get(GlobalId.MODE_ON_HOOK_SPAWN_NUM).args);

        //获取玩家所在当前地图可生成的宠物
        ModeOnHook info = this.getInfo(player).data;
        TbGOnHookMaps mapInfo = TD.DATA.getTbGOnHookMaps().get(info.getOnHookMap());
        List<TbGRole> tdPets = Arrays.stream(mapInfo.petIds).mapToObj(petId -> TD.DATA.getTbGRole().get(petId)).collect(Collectors.toList());

        //如果小于最大数量则生成
        while (creeps.size() < max){

            //目前随机生成宠物
            GOnHookMessage.GOnHookPet hookPet = GOnHookMessage.GOnHookPet.newBuilder()
                    //生成唯一Id
                    .setKey(RandomUtil.randomString(8))
                    //随机取配置表宠物中的宠物
                    .setPetTbId(tdPets.get(RandomUtil.randomInt(0, tdPets.size())).id)
                    .build();

            //添加到宠物列表中
            creeps.put(hookPet.getKey(),hookPet);

        }

        //返回生成好的宠物
        return ResponseEntity.ok().headers(headers).body(GOnHookMessage.GOnHookPets.newBuilder().addAllPets(creeps.values()).build().toByteArray());
    }

    //出售野怪
    @ApiImplicitParams({})
    @ApiOperation(value = "出售野怪")
    @PostMapping("/onSellCreeps/{creepId}")
    @Transactional
    public NewsResource<Boolean> onSellCreeps(@PathVariable String creepId, @CurrentPlayer Player player){

        //玩家的野怪
        Map<String, GOnHookMessage.GOnHookPet> creeps;
        //找到野怪删除
        if(Objects.isNull(creeps = CREEPS.get(player.getPlayerId()))) return NewsResource.onFail("卖出失败",false);

        GOnHookMessage.GOnHookPet onHookPet = creeps.remove(creepId);
        if(Objects.nonNull(onHookPet)){
            Resource res;
            //增加玩家金币 (默认添加100)
            try{
                res = resourceService.addResourceValue(player.getPlayerId(), ResourceId.Gold.id, 100L);
            }catch (RetryException e){
                //出售失败
                creeps.put(creepId,onHookPet); //放回去
                throw new RetryException("");
            }
            return NewsResource.onSuccess(
                    "卖出成功",
                    true,
                    ResourceUpdateOV.onUpdate(res)  //更新金币
                );
        }else{
            return NewsResource.onFail("卖出失败",false);
        }

    }

    //捕捉野怪
    @ApiImplicitParams({})
    @ApiOperation(value = "捕捉野怪")
    @PostMapping("/onCatchCreeps/{creepId}")
    public NewsContext<PlayerPet> onCatchCreeps(@PathVariable String creepId,@CurrentPlayer Player player){

        //(目前测试阶段 捕捉概率 100%)

        //获取玩家生成的野怪
        Map<String, GOnHookMessage.GOnHookPet> creeps = CREEPS.get(player.getPlayerId());

        //如果没有则生成类
        if(Objects.isNull(creeps)){
            CREEPS.put(player.getPlayerId(),creeps = new HashMap<>());
            return NewsContext.onFail("捕捉失败");
        }

        GOnHookMessage.GOnHookPet onHookPet;
        if(Objects.isNull(onHookPet = creeps.remove(creepId))){
            return NewsContext.onFail("捕捉失败");
        }

        //捕捉
        PlayerPet pet = PlayerPet.builder().petPlayerId(player.getPlayerId()).petTbId(onHookPet.getPetTbId()).petLevel(0).build();
        if(playerPetService.save(pet)){
            return NewsContext.onSuccess("捕捉成功",pet);
        }else{
            return NewsContext.onFail("捕捉失败");
        }

    }

}
