package cn.jisol.game.network;

import cn.jisol.game.actions.GActionEnum;
import cn.jisol.game.actions.GRefereeAction;
import cn.jisol.game.controller.game.GPlayerController;
import cn.jisol.game.data.Cache;
import cn.jisol.game.network.client.GClient;
import cn.jisol.game.network.client.GRefereeClient;
import cn.jisol.ngame.network.JNetwork;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Controller;

import javax.websocket.OnClose;
import javax.websocket.OnMessage;
import javax.websocket.OnOpen;
import javax.websocket.Session;
import javax.websocket.server.ServerEndpoint;
import java.io.InputStream;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.ConcurrentHashMap;

@ServerEndpoint(
    value = "/websocket/{token}"
)
@Controller
public class WebSocket {

    static GPlayerController playerController;

    @Autowired
    private void setPlayerController(GPlayerController playerController){
        WebSocket.playerController = playerController;
    }

    //玩家列表
    public static final Map<String, GClient> CLIENTS = new ConcurrentHashMap<>();
    //裁决员列表
    public static final Map<String, GRefereeClient> REFEREES = new ConcurrentHashMap<>();

    static {

        //将裁决员给裁决行为类
        GRefereeAction.REFEREES = REFEREES;

    }

    @OnOpen
    public void onOpen(Session session){

        String token = session.getPathParameters().get("token");
        GClient client = new GClient(token,session);

        if(Objects.isNull(client.user)){
            //发送Token过期请求
            client.invoke(GActionEnum.TOKEN_EXPIRED);
            //关闭连接
            client.Close();
            return;
        }

        client.player = playerController.getPlayerInfo(client.user).data;

        //判断玩家是否是裁决员是的话添加到裁决员列表中
        if(Cache.REFEREES.contains(client.user.getUserId()+"")){
            REFEREES.put(session.getId(),new GRefereeClient(token,session));
            System.out.printf("[WebSocket] %s 裁决员连接成功.\n",session.getId());
        }else{
            CLIENTS.put(session.getId(),client);
            System.out.printf("[WebSocket] %s 连接成功.\n",session.getId());
        }

    }

    @OnMessage
    public void onMessage(Session session, InputStream inputStream){
        GClient client;
        if(Objects.isNull(client = CLIENTS.get(session.getId())))
            client = REFEREES.get(session.getId());
        JNetwork.onMessage(inputStream,client,client,CLIENTS,client.user,client.player);
    }

    @OnClose
    public void onClose(Session session){
        CLIENTS.remove(session.getId());
        REFEREES.remove(session.getId());
    }

}
