package cn.jisol.game.service.impl;

import cn.jisol.game.controller.exception.ResourceNotException;
import cn.jisol.game.data.ResourceId;
import cn.jisol.game.entity.game.Player;
import cn.jisol.game.entity.game.Resource;
import cn.jisol.game.mapper.ResourceMapper;
import cn.jisol.game.service.ResourceService;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import org.springframework.retry.RetryException;
import org.springframework.retry.annotation.Retryable;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.Objects;

@Service
public class ResourceServiceImpl extends ServiceImpl<ResourceMapper, Resource> implements ResourceService {

    /**
     * @param playerId 玩家Id
     * @param type  资源类型
     * @param value 添加的数量
     * @return
     */
    @Override
    public Resource addResourceValue(Long playerId, int type, Long value) {


        LambdaUpdateWrapper<Resource> eq = Wrappers.lambdaUpdate(Resource.class)
                .eq(Resource::getPlayerId, playerId) //玩家
                .eq(Resource::getResourceTbId, type);//资源

        if(value < 0) {
            eq.gt(Resource::getResourceValue,Math.abs(value));
        }

        //执行SQL
        boolean update = this.update(
                eq.setSql("version = version + 1,resource_value = resource_value + " + value)
        );

        //查询货币数量
        Resource res = this.getOne(
                Wrappers.lambdaQuery(Resource.class)
                        .eq(Resource::getPlayerId, playerId) //玩家
                        .eq(Resource::getResourceTbId, type) //资源
        );

        if(!update){
            if(Objects.isNull(res)){
                //没有资源信息 创建初始资源
                res = this.initResource(playerId,type);
                //并且重试
                throw new RetryException("");
            }else{
                //更新失败 材料不够
                throw new ResourceNotException("资源不够");
            }
        }


        //如果资源不够则报错
        if(res.getResourceValue() < 0) throw new ResourceNotException("资源不够");

        return res;

    }

    //初始化某个资源
    @Override
    public Resource initResource(Long playerId, int type) {
        Resource resource = Resource.builder().playerId(playerId).resourceTbId(type).resourceValue(0L).build();
        this.initResource(resource);
        return resource;
    }

    @Override
    public Resource initResource(Resource resource) {
        this.save(resource);
        return resource;
    }


}
